"""Main moe package."""

import logging

import pluggy

hookspec = pluggy.HookspecMarker("moe")
hookimpl = pluggy.HookimplMarker("moe")

log = logging.getLogger("moe")
log_fmt = "{levelname}: {message}"
logging.basicConfig(level=logging.WARNING, format=log_fmt, style="{")

# always set external loggers to warning
for key in logging.Logger.manager.loggerDict:
    if "moe" not in key:
        logging.getLogger(key).setLevel(logging.WARNING)



"""new field: country.

Revision ID: 30668259fcd6
Revises: 60608d9d2908
Create Date: 2022-10-06 20:35:34.296797

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "30668259fcd6"
down_revision = "60608d9d2908"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.add_column(sa.Column("country", sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.drop_column("country")



"""new field: og_date.

Revision ID: db3a470892c6
Revises: 817440447857
Create Date: 2022-10-08 06:20:42.841502

"""
import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision = "db3a470892c6"
down_revision = "817440447857"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.add_column(sa.Column("original_date", sa.Date(), nullable=True))


def downgrade():
    with op.batch_alter_table("album", schema=None) as batch_op:
        batch_op.drop_column("original_date")



"""Edits music in the library."""

import pluggy

import moe
from moe import config

from . import edit_cli, edit_core
from .edit_core import *

__all__ = []
__all__.extend(edit_core.__all__)


@moe.hookimpl
def plugin_registration():
    """Only register the cli sub-plugin if the cli is enabled."""
    config.CONFIG.pm.register(edit_core, "edit_core")
    if config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.register(edit_cli, "edit_cli")



"""Import prompt."""

import functools
import logging
from typing import Optional, Union

import pluggy
from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

import moe
import moe.cli
from moe import config
from moe.cli import console
from moe.library import Album, Track
from moe.plugins.moe_import.import_core import CandidateAlbum
from moe.util.cli import PromptChoice, choice_prompt
from moe.util.core import get_matching_tracks

log = logging.getLogger("moe.cli.import")

__all__ = ["candidate_prompt", "import_prompt"]


class AbortImport(Exception):
    """Used to abort the import process."""


class Hooks:
    """Import plugin cli hook specifications."""

    @staticmethod
    @moe.hookspec
    def add_candidate_prompt_choice(prompt_choices: list[PromptChoice]):
        """Add a user input choice to the candidate prompt.

        ``func`` will be supplied the following keyword arguments:

            * ``new_album``: Album being added to the library.
            * ``candidates``: Candidate albums with all the import changes.

        Args:
            prompt_choices: List of prompt choices. To add a prompt choice, simply
                append it to this list.

        Example:
            .. code:: python

                prompt_choices.append(
                    PromptChoice(
                        title="Abort", shortcut_key="x", func=_abort_changes
                    )
                )
        """

    @staticmethod
    @moe.hookspec
    def add_import_prompt_choice(prompt_choices: list[PromptChoice]):
        """Add a user input choice to the import prompt.

        ``func`` will be supplied the following keyword arguments:

            * ``new_album``: Album being added to the library.
            * ``candidate``: Candidate album with all the import changes.

        Args:
            prompt_choices: List of prompt choices. To add a prompt choice, simply
                append it to this list.

        Example:
            .. code:: python

                prompt_choices.append(
                    PromptChoice(
                        title="Abort", shortcut_key="x", func=_abort_changes
                    )
                )
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `import` cli hookspecs to Moe."""
    from moe.plugins.moe_import.import_cli import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl
def add_candidate_prompt_choice(prompt_choices: list[PromptChoice]):
    """Adds the ``abort`` prompt choice to the user prompt."""
    prompt_choices.append(
        PromptChoice(title="Abort", shortcut_key="x", func=_abort_changes)
    )


@moe.hookimpl
def add_import_prompt_choice(prompt_choices: list[PromptChoice]):
    """Adds the ``apply`` and ``abort`` prompt choices to the user prompt."""
    prompt_choices.append(
        PromptChoice(title="Apply changes", shortcut_key="a", func=_apply_changes)
    )
    prompt_choices.append(
        PromptChoice(title="Abort", shortcut_key="x", func=_abort_changes)
    )


@moe.hookimpl
def process_candidates(new_album: Album, candidates: list[CandidateAlbum]):
    """Use the import prompt to select and process the imported candidate albums."""
    if candidates:
        try:
            candidate_prompt(new_album, candidates[:5])
        except AbortImport as err:
            log.debug(err)
            raise SystemExit(0) from err


def candidate_prompt(new_album: Album, candidates: list[CandidateAlbum]):
    """Runs the interactive prompt for a user to select a candidate to import.

    Args:
        new_album: Album being added to the library.
        candidates: List of candidates to choose from.

    Raises:
        AbortImport: Import prompt was aborted by the user.
    """
    prompt_choices: list[PromptChoice] = []

    for num, candidate in enumerate(candidates, start=1):
        prompt_choices.append(
            PromptChoice(
                _fmt_candidate_info(candidate),
                str(num),
                functools.partial(_select_candidate, candidate_num=num - 1),
            )
        )
    config.CONFIG.pm.hook.add_candidate_prompt_choice(prompt_choices=prompt_choices)

    prompt_choice = choice_prompt(
        prompt_choices, question="Which album would you like to import?"
    )
    prompt_choice.func(new_album, candidates)


def _fmt_candidate_info(candidate: CandidateAlbum) -> str:
    """Formats a candidates info for the candidate prompt."""
    sub_header_values = []
    for str_field in ["media", "country", "label"]:
        if value := getattr(candidate.album, str_field):
            sub_header_values.append(value)
    sub_header_values.extend(candidate.sub_header_info)
    sub_header = " | ".join(sub_header_values)

    return (
        str(candidate)
        + "\n"
        + " " * (9 + len(candidate.match_value_pct))
        + sub_header
        + "\n"
        + " " * (9 + len(candidate.match_value_pct))
        + candidate.source_str
        + "\n"
    )


def _select_candidate(
    new_album: Album, candidates: list[CandidateAlbum], candidate_num: int
):
    """Runs the import prompt for a selected candidate."""
    import_prompt(new_album, candidates[candidate_num])


def import_prompt(
    new_album: Album,
    candidate: CandidateAlbum,
):
    """Runs the interactive prompt for the given album changes.

    Args:
        new_album: Album being added to the library. Any changes will be applied to
            this album.
        candidate: New candidate album with all metadata changes. Will be compared
            against ``old_album``.

    Raises:
        AbortImport: Import prompt was aborted by the user.
    """
    log.debug(f"Running import prompt. [{new_album=!r}, {candidate=!r}]")

    console.print(_fmt_import_updates(new_album, candidate))

    prompt_choices: list[PromptChoice] = []
    config.CONFIG.pm.hook.add_import_prompt_choice(prompt_choices=prompt_choices)

    prompt_choice = choice_prompt(prompt_choices)
    prompt_choice.func(new_album, candidate)


def _apply_changes(
    new_album: Album,
    candidate: CandidateAlbum,
):
    """Applies the album changes."""
    log.debug("Applying changes from import prompt.")

    for old_track, new_track in get_matching_tracks(new_album, candidate.album):
        if not old_track and new_track:
            candidate.album.tracks.remove(new_track)  # missing track
        elif old_track and not new_track:
            new_album.tracks.remove(old_track)  # unmatched track
        elif (
            old_track
            and new_track
            and new_album.get_track(new_track.track_num, new_track.disc) != old_track
        ):
            # matchup track and disc numbers of matches to ensure they merge properly
            old_track.track_num = new_track.track_num
            old_track.disc = new_track.disc

    new_album.merge(candidate.album, overwrite=True)


def _abort_changes(
    new_album: Album,
    candidate: CandidateAlbum,
):
    """Aborts the album changes."""
    raise AbortImport("Import prompt aborted; no changes made.")


def _fmt_import_updates(new_album: Album, candidate: CandidateAlbum) -> Panel:
    """Formats import updates between `new_album` and the `candidate`."""
    album_text = _fmt_album(new_album, candidate)
    track_text = _fmt_tracks(new_album, candidate)

    return Panel(
        Group(album_text, track_text),
        title="Import Updates",
        border_style="light_cyan1",
    )


def _fmt_album(new_album: Album, candidate: CandidateAlbum) -> Text:
    """Formats the header for the import panel."""
    header_text = Text(justify="center", style="bold")

    for header_field in ("title", "artist"):
        field_changes = _fmt_field_changes(new_album, candidate.album, header_field)
        if not field_changes:
            field_changes = Text(getattr(new_album, header_field))

        header_text.append_text(field_changes).append("\n")

    sub_header_text = Text()
    for sub_header in ("media", "country", "label"):
        field_changes = _fmt_field_changes(new_album, candidate.album, sub_header)
        if not field_changes:
            if getattr(new_album, sub_header):
                field_changes = Text(getattr(new_album, sub_header))

        if sub_header_text and field_changes:
            sub_header_text.append(" | ")
        if field_changes:
            sub_header_text.append_text(field_changes)

    return (
        header_text.append_text(sub_header_text)
        .append("\n")
        .append(candidate.source_str)
        .append("\n")
    )


def _fmt_tracks(new_album: Album, candidate: CandidateAlbum) -> Table:
    """Formats the tracklist for the import panel."""
    track_table = Table(box=box.SIMPLE)
    track_table.add_column("status")
    track_table.add_column("disc")
    track_table.add_column("#")
    track_table.add_column("Artist")
    track_table.add_column("Title")

    matches = get_matching_tracks(new_album, candidate.album)
    matches.sort(
        key=lambda match: (
            getattr(match[1], "disc", 0),
            getattr(match[1], "track_num", 0),
        )
    )  # sort by new track's disc then track number

    unmatched_tracks: list[Track] = []
    missing_tracks: list[Track] = []
    for old_track, new_track in matches:
        if old_track and new_track:
            track_table.add_row(
                Text("matched", style="green"),
                _fmt_field_changes(old_track, new_track, "disc"),
                _fmt_field_changes(old_track, new_track, "track_num"),
                _fmt_field_changes(old_track, new_track, "artist"),
                _fmt_field_changes(old_track, new_track, "title"),
            )
        elif old_track and not new_track:
            unmatched_tracks.append(old_track)
        elif not old_track and new_track:
            missing_tracks.append(new_track)
    track_table.rows[-1].end_section = True

    for missing_track in sorted(missing_tracks):
        track_table.add_row(
            Text("missing", style="red"),
            str(missing_track.disc),
            str(missing_track.track_num),
            missing_track.artist,
            missing_track.title,
        )
    for unmatched_track in sorted(unmatched_tracks):
        track_table.add_row(
            Text("unmatched", style="red"),
            str(unmatched_track.disc),
            str(unmatched_track.track_num),
            unmatched_track.artist,
            unmatched_track.title,
        )

    return track_table


def _fmt_field_changes(
    old_item: Union[Album, Track], new_item: Union[Album, Track], field: str
) -> Optional[Text]:
    """Formats changes of a single field.

    Args:
        old_item: Old track or album to compare.
        new_item: New track or album to compare.
        field: Field to compare.

    Returns:
        A rich 'Text' based on the following cases:

        1. `old_item.field` DNE and `new_item.field` exists:
        `new_item.field` will be returned in green
        2. `old_item.field` exists and `new_item.field` DNE:
        `old_item.field`  will be striketroughed and in red
        3. `old_item.field` differs from `new_item.field`:
            "{old_item.field} -> {new_item.field}" and the arrow is yellow
        4. `old_item.field` is equal to `new_item.field`:
           `old_item.field` returned
        5. None if `old_item.field` and `new_item.field` DNE
    """
    old_field = getattr(old_item, field)
    new_field = getattr(new_item, field)

    if not old_field and new_field:
        return Text(str(new_field), style="green")
    if old_field and not new_field:
        return Text(str(old_field), style="red strike")
    if old_field != new_field:
        return (
            Text(str(old_field))
            .append(Text(" -> ", style="yellow"))
            .append(Text(str(new_field)))
        )

    if old_field:
        return Text(str(old_field))
    else:
        return None



"""User configuration of moe.

Each instance of Moe should only create a single `Config` object. This object will
contain all user settings and other related information and is set upon initializing
the configuration for the first time i.e. calling `Config()`. Once initialized, the
config can be accessed through import and should be treated as a constant::

    from moe import config
    print(config.CONFIG.settings.library_path)
"""

import importlib
import importlib.metadata
import importlib.util
import logging
import os
import re
import sys
from pathlib import Path
from types import ModuleType
from typing import NamedTuple, Optional, Union, cast

import dynaconf
import pluggy
import sqlalchemy
import sqlalchemy.event
import sqlalchemy.orm

import alembic.command
import alembic.config
import moe

session_factory = sqlalchemy.orm.sessionmaker(autoflush=False)
MoeSession = sqlalchemy.orm.scoped_session(session_factory)

__all__ = ["CONFIG", "Config", "ConfigValidationError", "ExtraPlugin"]

log = logging.getLogger("moe.config")

DEFAULT_PLUGINS = {
    "add",
    "cli",
    "duplicate",
    "edit",
    "import",
    "list",
    "move",
    "musicbrainz",
    "sync",
    "remove",
    "write",
}
CORE_PLUGINS = {
    "config": "moe.config",
    "album": "moe.library.album",
    "extra": "moe.library.extra",
    "track": "moe.library.track",
    "lib_item": "moe.library.lib_item",
}  # {name: module} of plugins that cannot be overwritten by the config

CONFIG = cast("Config", None)


class ConfigValidationError(Exception):
    """Error in the user's configuration."""


class Hooks:
    """Config hook specifications."""

    @staticmethod
    @moe.hookspec
    def add_config_validator(settings: dynaconf.base.LazySettings):
        """Add a settings validator for the configuration file.

        Args:
            settings: Moe's settings.

        Example:
            .. code:: python

                settings.validators.register(
                    dynaconf.Validator("MOVE.ASCIIFY_PATHS", must_exist=True)
                )

        See Also:
            https://www.dynaconf.com/validation/#validation for more info.
        """

    @staticmethod
    @moe.hookspec
    def add_hooks(pm: pluggy.manager.PluginManager):
        """Add hookspecs to be registered to Moe.

        Args:
            pm: PluginManager that registers the hookspec.

        Example:
            .. code:: python

                from moe.plugins.add import Hooks
                pm.add_hookspecs(Hooks)
        """

    @staticmethod
    @moe.hookspec
    def plugin_registration():
        """Allows actions after the initial plugin registration.

        In order for a module to implement and register plugin hooks, it must be
        registered as a separate plugin with the ``pm``. A plugin can be
        either just a module, or a full package.

        If a plugin is a package, only it's ``__init__.py`` will be initially
        registered, meaning only ``__init__.py`` will be able to run hook
        implementations at start-up. This hook is provided so each plugin can register
        it's individual sub-modules as appropriate.

        Important:
            Ensure any sub-modules you register as plugins are registered with the
            original plugin name as the prefix. This helps prevent naming conflicts.

        For example, see how the ``edit`` plugin conditionally enables its cli
        sub-module::

            @moe.hookimpl
            def plugin_registration():
                if config.CONFIG.pm.has_plugin("cli"):
                    config.CONFIG.pm.register(edit_cli, "edit_cli")

        This hook can also be used as a way of checking for plugin dependencies by
        inspecting the enabled plugins in the configuration.

        For example, because the ``list`` plugin only exists as a cli plugin, it will
        un-register itself and log a warning if the cli plugin is not enabled::

            @moe.hookimpl
            def plugin_registration():
                if not config.CONFIG.pm.has_plugin("cli"):
                    config.CONFIG.pm.set_blocked("list")
                    log.warning("You can't list stuff without a cli!")

        See Also:
            `pluggy.PluginManager documentation <https://pluggy.readthedocs.io/en/latest/api_reference.html>`_
        """  # noqa: E501

    @staticmethod
    @moe.hookspec
    def register_sa_event_listeners(session: sqlalchemy.orm.Session):
        """Registers new sqlalchemy event listeners.

        Args:
            session: Session to attach the listener to.

        Important:
            This hooks is for Moe internal use only and should not be used by plugins.

        Example:
            .. code:: python

                sqlalchemy.event.listen(
                    session,
                    "before_flush",
                    _my_func,
                )

            Then you can define ``_my_func`` as such::

                def _my_func(
                    session: sqlalchemy.orm.Session,
                    flush_context: sqlalchemy.orm.UOWTransaction,
                    instances: Optional[Any],
                ):
                    print("we made it")

        See Also:
            `SQLAlchemy ORM even documentation <https://docs.sqlalchemy.org/en/14/orm/events.html#orm-events>`_
        """  # noqa: E501


@moe.hookimpl
def add_config_validator(settings: dynaconf.base.LazySettings):
    """Validate move plugin configuration settings."""
    validators = [
        dynaconf.Validator("DEFAULT_PLUGINS", default=DEFAULT_PLUGINS),
        dynaconf.Validator("DISABLE_PLUGINS", default=set()),
        dynaconf.Validator("ENABLE_PLUGINS", default=set()),
        dynaconf.Validator("LIBRARY_PATH", default="~/Music"),
        dynaconf.Validator("ORIGINAL_DATE", default=False),
    ]
    settings.validators.register(*validators)


class ExtraPlugin(NamedTuple):
    """Used to specify extra plugins when initializing the config.

    Attributes:
        plugin: This is the class or module of the plugin to register.
        name: Name to register the plugin under.
    """

    plugin: Union[type, ModuleType]
    name: str


class Config:
    """Initializes moe configuration settings and database.

    Attributes:
        config_dir (Path): Filesystem path of the configuration directory.
        config_file (Path): Filesystem path of the configuration settings file.
        engine (sa.engine.base.Engine): Database engine in use.
        pm (pluggy.manager.PluginManager): Plugin manager that handles plugin logic.
        settings (dynaconf.base.LazySettings): User configuration settings.
    """

    def __init__(
        self,
        config_dir: Path = Path.home() / ".config" / "moe",  # noqa: B008
        settings_filename: str = "config.toml",
        extra_plugins: Optional[list[ExtraPlugin]] = None,
        engine: Optional[sqlalchemy.engine.base.Engine] = None,
        init_db=True,
    ):
        """Initializes the plugin manager and configuration directory.

        Args:
            config_dir: Filesystem path of the configuration directory where the
                settings and database files will reside. The environment variable
                ``MOE_CONFIG_DIR`` has precedence in setting this.
            extra_plugins: Any extra plugins that should be enabled in addition to those
                specified in the configuration.
            settings_filename: Name of the configuration settings file.
            engine: sqlalchemy database engine to use. Defaults to a sqlite db located
                in the ``config_dir``.
            init_db: Whether or not to initialize the database.
        """
        global CONFIG
        CONFIG = self

        try:
            self.config_dir = Path(os.environ["MOE_CONFIG_DIR"])
        except KeyError:
            self.config_dir = config_dir

        self.config_dir.mkdir(parents=True, exist_ok=True)

        self.config_file = self.config_dir / settings_filename
        self._extra_plugins = extra_plugins or []
        self._read_config()

        self.engine = engine
        if init_db:
            self._init_db()

    def _init_db(self, create_tables: bool = True):
        """Initializes the database.

        Moe uses sqlite by default.

        Args:
            create_tables: Whether or not to create and update the db tables.
                If doing db migrations manually, e.g. in alembic, this should be False.
        """
        log.debug(f"Initializing database. [{create_tables=!r}]")

        db_path = self.config_dir / "library.db"

        if not self.engine:
            self.engine = sqlalchemy.create_engine("sqlite:///" + str(db_path))

        session_factory.configure(bind=self.engine)

        # create and update database tables
        if create_tables:
            config_path = Path(__file__)
            alembic_cfg = alembic.config.Config(
                str(config_path.parents[1] / "alembic" / "alembic.ini")
            )
            alembic_cfg.attributes["configure_logger"] = False
            with self.engine.begin() as connection:
                alembic_cfg.attributes["connection"] = connection
                alembic.command.upgrade(alembic_cfg, "head")

        self.pm.hook.register_sa_event_listeners(config=self, session=MoeSession())

        # create regular expression function for sqlite queries
        @sqlalchemy.event.listens_for(self.engine, "begin")
        def sqlite_engine_connect(conn):
            conn.connection.create_function("regexp", 2, _regexp, deterministic=True)

        def _regexp(pattern: str, col_value) -> bool:
            """Use the python re module for sqlite regular expression functionality.

            Args:
                pattern: Regular expression pattern.
                col_value: Column value to match against. The match will be against
                    the str of the value.

            Returns:
                Whether or not the match was successful.
            """
            return re.search(pattern, str(col_value), re.IGNORECASE) is not None

        log.debug(f"Initialized database. [engine={self.engine!r}]")

    def _read_config(self):
        """Reads the user configuration settings.

        Searches for a configuration file at `config_dir / "config.toml"`.

        Raises:
            ConfigValidationError: Unable to parse the configuration file.
        """
        from moe.library.lib_item import PathType

        log.debug(f"Reading configuration file. [config_file={self.config_file}]")

        self.config_file.touch(exist_ok=True)

        self.settings = dynaconf.Dynaconf(
            envvar_prefix="MOE",  # export envvars with `export MOE_FOO=bar`
            settings_file=str(self.config_file.resolve()),
        )

        self._setup_plugins()
        self.pm.hook.add_config_validator(settings=self.settings)
        self._validate_settings()

        PathType.library_path = Path(self.settings.library_path)

    def _validate_settings(self):
        """Validates the given user configuration.

        Raises:
            ConfigValidationError: Error validating the configuration.
        """
        log.debug("Validating configuration settings.")

        try:
            self.settings.validators.validate()
        except dynaconf.validator.ValidationError as err:
            raise ConfigValidationError(err) from err

    def _setup_plugins(self, core_plugins: dict[str, str] = CORE_PLUGINS):
        """Setup pm and hook logic.

        Args:
            core_plugins: Optional mapping of core plugin modules to names.
                These plugins cannot be overwritten by the user configuration.
        """
        log.debug("Setting up plugins.")

        self.pm = pluggy.PluginManager("moe")

        # register core modules that are not considered plugins
        for plugin_name, module in core_plugins.items():
            self.pm.register(importlib.import_module(module), plugin_name)

        # need to validate `config` specific settings separately so we have access to
        # the 'default_plugins' setting
        self.pm.add_hookspecs(Hooks)
        self.pm.hook.add_config_validator(settings=self.settings)
        self._validate_settings()

        config_plugins = (
            set(self.settings.default_plugins) | set(self.settings.enable_plugins)
        ) - set(self.settings.disable_plugins)

        # the 'import' plugin maps to the 'moe_import' package
        if "import" in config_plugins:
            config_plugins.remove("import")
            config_plugins.add("moe_import")

        if "cli" in config_plugins:
            self.pm.register(importlib.import_module("moe.cli"), name="cli")

        # register plugin hookimpls for all enabled plugins
        internal_plugin_dir = Path(__file__).resolve().parent / "plugins"
        self._register_local_plugins(
            config_plugins, internal_plugin_dir, "moe.plugins."
        )
        if Path(self.config_dir / "plugins").exists():
            sys.path.append(str(self.config_dir / "plugins"))
            self._register_local_plugins(config_plugins, self.config_dir / "plugins")

        # register all third-party installed plugins
        plugins = importlib.metadata.entry_points().get("moe.plugins")
        if plugins:
            for plugin in plugins:
                if plugin.name in config_plugins:
                    plugin.load()
                    self.pm.register(plugin.load(), plugin.name)

        # register plugin hookimpls for all extra plugins
        for extra_plugin in self._extra_plugins:
            self.pm.register(extra_plugin.plugin, extra_plugin.name)

        # register individual plugin sub-modules
        self.pm.hook.plugin_registration(pm=self.pm)

        # register plugin hookspecs for all enabled plugins
        self.pm.hook.add_hooks(pm=self.pm)

        log.debug(f"Registered plugins. [plugins={self.pm.list_name_plugin()}]")

    def _register_local_plugins(
        self, enabled_plugins: set[str], plugin_dir: Path, pkg_name: str = ""
    ):
        """Registers all internal plugins in `enabled_plugins`.

        Args:
            enabled_plugins: All enabled plugins as specified by the config.
            plugin_dir: Directory of plugins to register.
            pkg_name: Optional common package name the plugins belong to.
                Include the trailing '.'.
        """
        for plugin_path in plugin_dir.iterdir():
            plugin_name = plugin_path.stem

            if plugin_path.stem in enabled_plugins:
                plugin = importlib.import_module(pkg_name + plugin_name)
                self.pm.register(plugin, plugin_name)



"""Adds the ``move`` command to the cli."""

import argparse
import logging
from typing import cast

import sqlalchemy.orm

import moe
from moe.library import Album
from moe.plugins import move as moe_move
from moe.util.cli import cli_query

log = logging.getLogger("moe.cli.move")


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``move`` command to Moe's CLI."""
    move_parser = cmd_parsers.add_parser(
        "move",
        aliases=["mv"],
        description="Moves items in the library according to the user configuration.",
        help="move items in the library",
    )
    move_parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="display what will be moved without actually moving the items",
    )
    move_parser.set_defaults(func=_parse_args)


def _parse_args(args: argparse.Namespace):
    """Parses the given commandline arguments.

    Items will be moved according to the given user configuration.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Invalid query or no items found to move.
    """
    albums = cast(list[Album], cli_query("*", query_type="album"))

    if args.dry_run:
        dry_run_str = _dry_run(albums)
        if dry_run_str:
            print(dry_run_str.lstrip())
    else:
        for album in albums:
            moe_move.move_item(album)


def _dry_run(albums: list[Album]) -> str:
    """Returns a string of output representing a 'dry-run' of moving albums."""
    dry_run_str = ""

    for dry_album in albums:
        album_dest = moe_move.fmt_item_path(dry_album)
        if album_dest != dry_album.path:
            dry_run_str += f"\n{dry_album.path}\n\t-> {album_dest}"

        # temporarily set the album's path so track/extra dests use the right
        # album directory
        sqlalchemy.orm.attributes.set_committed_value(dry_album, "path", album_dest)

        for dry_track in dry_album.tracks:
            track_dest = moe_move.fmt_item_path(dry_track)
            if track_dest != dry_track.path:
                dry_run_str += f"\n{dry_track.path}\n\t-> {track_dest}"
        for dry_extra in dry_album.extras:
            extra_dest = moe_move.fmt_item_path(dry_extra)
            if extra_dest != dry_extra.path:
                dry_run_str += f"\n{dry_extra.path}\n\t-> {extra_dest}"

    return dry_run_str



"""Core api for importing albums."""

import itertools
import logging
import operator
from dataclasses import dataclass, field

import pluggy

import moe
from moe import config
from moe.library import Album, LibItem, Track

__all__ = ["CandidateAlbum", "import_album"]

log = logging.getLogger("moe.import")


@dataclass
class CandidateAlbum:
    """A single candidate for the import process.

    Attributes:
        album (Album): The candidate album.
        match_value (float): 0 to 1 scale of how well the candidate album matches with
            the album being imported.
        source_str (str): A string identifying the release and it's source
            e.g. 'Musicbrainz: 1234'. This will be displayed as the last line in the
            candidate prompt.
        sub_header_info (list[str]): List of any additional info to include in the
            candidate sub-header. The following fields are already included:
            ['media', 'country', 'label']
        match_value_pct (str): ``match_value`` as a percentage.
    """

    album: Album
    match_value: float
    source_str: str
    sub_header_info: list[str] = field(default_factory=list)

    @property
    def match_value_pct(self) -> str:
        """Formats `match_value` as a percentage."""
        return f"{round(self.match_value * 100, 1)}%"

    def __str__(self):
        """String representation of a CandidateAlbum."""
        return f"[{self.match_value_pct}] {self.album.artist} - {self.album.title}"


class Hooks:
    """Import core plugin hook specifications."""

    @staticmethod
    @moe.hookspec
    def get_candidates(album: Album) -> list[CandidateAlbum]:  # type: ignore
        """Return candidate albums from implemented sources based on the given album.

        This hook should be used to import metadata from an external source and return
        any candidate albums to import from. The candidate albums will then be processed
        in the :meth:`process_candidates` hook.

        Note:
            This hook runs within the :meth:`pre_add` hook.

        Args:
            album: Album being added to the library.

        Returns:
            New candidate album.
        """  # noqa: DAR202

    @staticmethod
    @moe.hookspec
    def process_candidates(new_album: Album, candidates: list[CandidateAlbum]):
        """Process the imported candidate albums.

        If you wish to save and apply any candidate album metadata, it should be applied
        onto the original album, ``new_album``.

        Ensure any potential conflicts with existing albums in the database are
        resolved.

        Args:
            new_album: Album being added to the library.
            candidates: New candidate albums with imported metadata sorted by how well
                 they match ``new_album``.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `import` core hookspecs to Moe."""
    from moe.plugins.moe_import.import_core import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl
def pre_add(item: LibItem):
    """Fixes album metadata via external sources prior to it being added to the lib."""
    if isinstance(item, Album):
        album = item
    elif isinstance(item, Track):
        album = item.album_obj
    else:
        return

    import_album(album)


def import_album(album: Album):
    """Imports album metadata for an album."""
    log.debug(f"Importing album metadata. [{album=!r}]")

    candidates = config.CONFIG.pm.hook.get_candidates(album=album)
    candidates = list(itertools.chain.from_iterable(candidates))
    candidates.sort(key=operator.attrgetter("match_value"), reverse=True)

    config.CONFIG.pm.hook.process_candidates(
        new_album=album,
        candidates=candidates,
    )

    log.debug(f"Imported album metadata. [{album=!r}]")



"""Adds music to the library."""

import moe
from moe import config

from . import add_cli, add_core
from .add_core import *

__all__ = []
__all__.extend(add_core.__all__)


@moe.hookimpl
def plugin_registration():
    """Only register the cli sub-plugin if the cli is enabled."""
    config.CONFIG.pm.register(add_core, "add_core")
    if config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.register(add_cli, "add_cli")



"""All logic regarding matching albums and tracks against each other."""

import difflib
import logging
from typing import Optional, Union

from moe.library import Album, Track

log = logging.getLogger("moe")

__all__ = ["get_match_value", "get_matching_tracks"]

TrackMatch = tuple[Optional[Track], Optional[Track]]
TrackCoord = tuple[
    tuple[int, int], tuple[int, int]
]  # ((a.disc, a.track_num), (b.disc, b.track_num))

MATCH_ALBUM_FIELD_WEIGHTS = {
    "artist": 0.8,
    "country": 0.3,
    "date": 0.2,
    "disc_total": 0.8,
    "label": 0.1,
    "media": 0.4,
    "title": 0.9,
}  # how much to weigh matches of various fields

MATCH_TRACK_FIELD_WEIGHTS = {
    "disc": 0.3,
    "title": 0.7,
    "track_num": 0.9,
}  # how much to weigh matches of various fields


def get_match_value(item_a: Union[Album, Track], item_b: Union[Album, Track]) -> float:
    """Returns a similarity "match value" between two albums on a scale of 0 to 1.

    Args:
        item_a: First item to compare.
        item_b: Second item to compare.

    Returns:
        The match value is a weighted sum according to the defined weights for each
        applicable field.
    """
    log.debug(f"Determining match value between items. [{item_a=!r}, {item_b=!r}]")

    if type(item_a) is not type(item_b):
        return 0

    if isinstance(item_a, Album):
        field_weights = MATCH_ALBUM_FIELD_WEIGHTS
    elif isinstance(item_a, Track):
        field_weights = MATCH_TRACK_FIELD_WEIGHTS

    penalties = []
    for field, weight in field_weights.items():
        value_a = getattr(item_a, field)
        value_b = getattr(item_b, field)

        if not value_a and not value_b:
            penalty = 0.1
        elif (value_a and not value_b) or (value_b and not value_a):
            penalty = 0.2
        else:
            if isinstance(value_a, str):
                penalty = 1 - difflib.SequenceMatcher(None, value_a, value_b).ratio()
            else:
                if value_a != value_b:
                    penalty = 1
                else:
                    penalty = 0

        penalties.append(penalty * weight)

    match_value = 1 - sum(penalties) / sum(field_weights.values())

    log.debug(f"Determined match value between items. [{match_value=!r}]")
    return match_value


def get_matching_tracks(  # noqa: C901 (I don't see benefit from splitting)
    album_a: Album, album_b: Album, match_threshold: float = 0.7
) -> list[TrackMatch]:
    """Returns a list of tuples of track match pairs.

    Args:
        album_a: Album tracks will be matched against ``album_b``.
        album_b: Album tracks will be matched against ``album_a``.
        match_threshold: Threshold match value to consider a track a match with
            another. See ``get_track_match_value()`` for more info.

    Returns:
        A list of tuples of track match pairs. Each track in ``album_a`` and
        ``album_b`` will appear with its respective track match or ``None`` if no match
        was found. Each tuple represents a match and will be in the form
        ``(album_a_track, album_b_track)``.
    """
    log.debug(
        f"Finding matching tracks. [{album_a=!r}, {album_b=!r}, {match_threshold=!r}]"
    )

    # get all match values for every pair of tracks between both albums
    track_match_values: dict[TrackCoord, float] = {}
    for a_track in album_a.tracks:
        for b_track in album_b.tracks:
            track_match_values[
                ((a_track.disc, a_track.track_num), (b_track.disc, b_track.track_num))
            ] = get_match_value(a_track, b_track)

    # Greedy algorithm that assigns a match to each track pair in order of greatest
    # match value.
    #
    # Potential issues:
    # 1. Consider the following match_value matrix:
    #    a1 a2
    # b1 .8 .9
    # b2  0 .8
    # In this case, a2 will match with b1, while both a1 and b2 will be left with no
    # match. A better solution may be to have (a1, b1) and (a2, b2) as matches.
    #
    # Potential improvements:
    # 1. Solve for greatest sum of match values between all row/col pairs?
    #     a. Not sure if this would actually produce a more desireable solution.
    track_match_values = dict(
        sorted(track_match_values.items(), reverse=True, key=lambda match: match[1])
    )  # sort by match value (descending)
    track_matches: list[TrackMatch] = []  # [(a_track, b_track)]
    for track_pair, match_value in track_match_values.items():
        if match_value < match_threshold:
            continue

        track_matches.append(
            (
                album_a.get_track(track_num=track_pair[0][1], disc=track_pair[0][0]),
                album_b.get_track(track_num=track_pair[1][1], disc=track_pair[1][0]),
            )
        )

        # prevent subsequent matches
        for track_pair2 in track_match_values.keys():
            if track_pair[0] == track_pair2[0] or track_pair[1] == track_pair2[1]:
                track_match_values[track_pair2] = -1

    # pair unmatched tracks with `None`
    for a_track in album_a.tracks:
        if not any(a_track == track_match[0] for track_match in track_matches):
            track_matches.append((a_track, None))
    for b_track in album_b.tracks:
        if not any(b_track == track_match[1] for track_match in track_matches):
            track_matches.append((None, b_track))

    log.debug(f"Found matching tracks. [matches={track_matches!r}]")
    return track_matches



"""CLI-specific query help functionality."""

import argparse
import logging

from moe.library import LibItem
from moe.query import QueryError, query

__all__ = ["cli_query", "query_parser"]

log = logging.getLogger("moe.cli")

query_parser = argparse.ArgumentParser(
    add_help=False, formatter_class=argparse.RawTextHelpFormatter
)
"""Argument parser for a query.

Passing this parser as a parent to your command allows your command to accept queries.

Example:
    .. code:: python

        ls_parser = cmd_parsers.add_parser(
            "list",
            description="Lists music in the library.",
            help="list music in the library",
            parents=[query_parser],
        )

Given parsed commandline arguments as ``args``, you can then access the given
query string as ``args.query`` and the query type as ``args.query_type``. Both options
can be passed directly to :meth:`moe.query.query` or :meth:`moe.util.cli.cli_query`.

See Also:
    `argparse parents documentation
    <https://docs.python.org/3/library/argparse.html#parents>`_
"""

query_parser.add_argument("query", help="query the library for matching tracks")

query_type_group = query_parser.add_mutually_exclusive_group()
query_type_group.add_argument(
    "-a",
    "--album",
    action="store_const",
    const="album",
    dest="query_type",
    help="query for matching albums",
)
query_type_group.add_argument(
    "-e",
    "--extra",
    action="store_const",
    const="extra",
    dest="query_type",
    help="query for matching extras",
)
query_parser.set_defaults(query_type="track")


def cli_query(query_str: str, query_type: str) -> list[LibItem]:
    """Wrapper around the core query call, with some added cli error handling.

    Args:
        query_str: Query string to parse. See HELP_STR for more info.
        query_type: Type of library item to return: either 'album', 'extra', or 'track'.

    Returns:
        All items matching the given query found in ``args``.

    Raises:
        SystemExit: QueryError or no items returned from the query.
    """
    try:
        items = query(query_str, query_type)
    except QueryError as err:
        log.error(err)
        raise SystemExit(1) from err

    if not items:
        log.error("No items found for given query.")
        raise SystemExit(1)

    return items



"""Core api for moving items."""

import logging
import re
import shutil
from contextlib import suppress
from pathlib import Path
from typing import Callable, Union

import dynaconf
import pluggy
from unidecode import unidecode

import moe
from moe import config
from moe.library import Album, Extra, LibItem, Track

__all__ = ["copy_item", "fmt_item_path", "move_item"]

log = logging.getLogger("moe.move")


class Hooks:
    """Move plugin hook specifications."""

    @staticmethod
    @moe.hookspec
    def create_path_template_func() -> list[Callable]:  # type: ignore
        """Create a custom path template function.

        Any functions returned by this hook will be available to be called from within
        the path templates defined by the move plugin.

        Returns:
            A list of all custom path functions your plugin creates. The list should
            contain the callable functions themselves.
        """  # noqa: DAR202


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `add` hookspecs to Moe."""
    from moe.plugins.move.move_core import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl
def add_config_validator(settings: dynaconf.base.LazySettings):
    """Validate move plugin configuration settings."""
    default_album_path = "{album.artist}/{album.title} ({album.year})"
    default_extra_path = "{e_unique(extra)}"
    default_track_path = (
        "{f'Disc {track.disc:02}' if album.disc_total > 1 else ''}/"
        "{track.track_num:02} - {track.title}{track.path.suffix}"
    )

    validators = [
        dynaconf.Validator("MOVE.ASCIIFY_PATHS", default=False),
        dynaconf.Validator("MOVE.ALBUM_PATH", default=default_album_path),
        dynaconf.Validator("MOVE.EXTRA_PATH", default=default_extra_path),
        dynaconf.Validator("MOVE.TRACK_PATH", default=default_track_path),
    ]
    settings.validators.register(*validators)


@moe.hookimpl(trylast=True)
def edit_new_items(items: list[LibItem]):
    """Copies and formats the path of an item after it has been added to the library."""
    for item in items:
        copy_item(item)


@moe.hookimpl
def create_path_template_func() -> list[Callable]:
    """Adds custom functions for the path templates."""
    return [e_unique]


def e_unique(extra: Extra) -> str:
    """Returns a unique filename for an extra within its album."""
    extra_names = [album_extra.path.name for album_extra in extra.album_obj.extras]

    if (name_count := extra_names.count(extra.path.name)) > 1:
        return extra.path.stem + f" ({name_count - 1})" + extra.path.suffix

    return extra.path.name


########################################################################################
# Format paths
########################################################################################
def fmt_item_path(item: LibItem) -> Path:
    """Returns a formatted item path according to the user configuration."""
    log.debug(f"Formatting item path. [path={item.path}]")

    if isinstance(item, Album):
        new_path = _fmt_album_path(item)
    elif isinstance(item, Extra):
        new_path = _fmt_extra_path(item)
    elif isinstance(item, Track):
        new_path = _fmt_track_path(item)
    else:
        raise NotImplementedError

    if config.CONFIG.settings.move.asciify_paths:
        new_path = Path(unidecode(str(new_path)))

    log.debug(f"Formatted item path. [path={new_path}]")
    return new_path


def _fmt_album_path(album: Album) -> Path:
    """Returns a formatted album directory according to the user configuration."""
    library_path = Path(config.CONFIG.settings.library_path).expanduser()
    album_path = _eval_path_template(config.CONFIG.settings.move.album_path, album)

    return library_path / album_path


def _fmt_extra_path(extra: Extra) -> Path:
    """Returns a formatted extra path according to the user configuration."""
    album_path = _fmt_album_path(extra.album_obj)
    extra_path = _eval_path_template(config.CONFIG.settings.move.extra_path, extra)

    return album_path / extra_path


def _fmt_track_path(track: Track) -> Path:
    """Returns a formatted track path according to the user configuration."""
    album_path = _fmt_album_path(track.album_obj)
    track_path = _eval_path_template(config.CONFIG.settings.move.track_path, track)

    return album_path / track_path


def _eval_path_template(template, lib_item) -> str:
    """Evaluates and sanitizes a path template.

    Args:
        template: Path template.
            See `_lazy_fstr_item()` for more info on accepted f-string templates.
        lib_item: Library item associated with the template.

    Returns:
        Evaluated path.
    """
    template_parts = template.split("/")
    sanitized_parts = []
    for template_part in template_parts:
        path_part = _lazy_fstr_item(template_part, lib_item)
        sanitized_part = _sanitize_path_part(path_part)
        if sanitized_part:
            sanitized_parts.append(sanitized_part)

    return "/".join(sanitized_parts)


def _lazy_fstr_item(template: str, lib_item: LibItem) -> str:
    """Evalutes the given f-string template for a specific library item.

    Args:
        template: f-string template to evaluate.
            All library items should have their own template and refer to variables as:
                Album: album (e.g. {album.title}, {album.artist})
                Track: track (e.g. {track.title}, {track.artist})
                Extra: extra (e.g. {extra.path.name}
        lib_item: Library item referenced in the template.


    Example:
        The default path template for an album is::

            {album.artist}/{album.title} ({album.year})

    Returns:
        Evaluated f-string.

    Raises:
        NotImplementedError: You discovered a new library item!
    """
    # add the appropriate library item to the scope
    if isinstance(lib_item, Album):
        album = lib_item  # noqa: F841
    elif isinstance(lib_item, Track):
        track = lib_item  # noqa: F841
        album = track.album_obj
    elif isinstance(lib_item, Extra):
        extra = lib_item  # noqa: F841
        album = extra.album_obj  # noqa: F841
    else:
        raise NotImplementedError

    plugin_funcs = config.CONFIG.pm.hook.create_path_template_func()
    for funcs in plugin_funcs:
        for func in funcs:
            globals()[func.__name__] = func

    return eval(f'f"""{template}"""')


def _sanitize_path_part(path_part: str) -> str:
    """Sanitizes a part of a path to be compatible with most filesystems.

    Note:
        Only sub-paths of the library path will be affected.

    Args:
        path_part: Path part to sanitize. Must be a single 'part' of a path, i.e. no /
            separators.

    Returns:
        Path part with all the replacements applied.
    """
    PATH_REPLACE_CHARS = {
        r"^\.": "_",  # leading '.' (hidden files on Unix)
        r'[<>:"\?\*\|\\/]': "_",  # <, >, : , ", ?, *, |, \, / (Windows reserved chars)
        r"\.$": "_",  # trailing '.' (Windows restriction)
        r"\s+$": "",  # trailing whitespace (Windows restriction)
    }

    for regex, replacement in PATH_REPLACE_CHARS.items():
        path_part = re.sub(regex, replacement, path_part)

    return path_part


########################################################################################
# Copy
########################################################################################
def copy_item(item: LibItem):
    """Copies an item to a destination as determined by the user configuration.

    Overwrites any existing files. Will create the destination if it does not already
    exist.
    """
    if isinstance(item, Album):
        _copy_album(item)
    elif isinstance(item, (Extra, Track)):
        _copy_file_item(item)


def _copy_album(album: Album):
    """Copies an album to a destination as determined by the user configuration."""
    dest = fmt_item_path(album)

    log.debug(f"Copying album. [{dest=}, {album=!r}]")

    dest.mkdir(parents=True, exist_ok=True)
    album.path = dest

    for track in album.tracks:
        _copy_file_item(track)

    for extra in album.extras:
        _copy_file_item(extra)

    log.info(f"Album copied. [{dest=}, {album=!r}]")


def _copy_file_item(item: Union[Extra, Track]):
    """Copies an extra or track to a destination as determined by the user config."""
    dest = fmt_item_path(item)
    if dest.exists() and dest.samefile(item.path):
        item.path = dest
        return

    log.debug(f"Copying item. [{dest=}, {item=!r}]")

    dest.parent.mkdir(parents=True, exist_ok=True)
    shutil.copyfile(item.path, dest)

    item.path = dest

    log.info(f"Copied item. [{dest=}, {item=!r}]")


########################################################################################
# Move
########################################################################################
def move_item(item: LibItem):
    """Moves an item to a destination as determined by the user configuration.

    Overwrites any existing files. Will create the destination if it does not already
    exist.
    """
    if isinstance(item, Album):
        _move_album(item)
    elif isinstance(item, (Extra, Track)):
        _move_file_item(item)


def _move_album(album: Album):
    """Moves an album to a given destination.

    Note:
        Empty leftover directories will be removed.
    """
    dest = fmt_item_path(album)
    old_album_dir = album.path

    log.debug(f"Moving album. [{dest=}, {album=!r}]")

    dest.mkdir(parents=True, exist_ok=True)
    album.path = dest

    for track in album.tracks:
        _move_file_item(track)

    for extra in album.extras:
        _move_file_item(extra)

    # remove any empty leftover directories
    for old_child in old_album_dir.rglob("*"):
        with suppress(OSError):
            old_child.rmdir()
    with suppress(OSError):
        old_album_dir.rmdir()
    for old_parent in old_album_dir.parents:
        with suppress(OSError):
            old_parent.rmdir()

    log.info(f"Moved album. [{dest=}, {album=!r}]")


def _move_file_item(item: Union[Extra, Track]):
    """Moves an extra or track to a destination as determined by the user config."""
    dest = fmt_item_path(item)
    if dest.exists() and dest.samefile(item.path):
        item.path = dest
        return

    log.debug(f"Moving item. [{dest=}, {item=!r}]")

    dest.parent.mkdir(parents=True, exist_ok=True)
    item.path.replace(dest)

    item.path = dest

    log.info(f"Moved item. [{dest=}, {item=!r}]")



"""A Track in the database and any related logic."""

import logging
from pathlib import Path
from typing import Any, Optional, TypeVar, cast

import mediafile
import pluggy
from sqlalchemy import JSON, Column, Integer, String
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import relationship
from sqlalchemy.schema import ForeignKey, UniqueConstraint

import moe
from moe import config
from moe.library.album import Album
from moe.library.lib_item import LibItem, LibraryError, PathType, SABase

__all__ = ["Track", "TrackError"]

log = logging.getLogger("moe.track")


class Hooks:
    """Track hook specifications."""

    @staticmethod
    @moe.hookspec
    def create_custom_track_fields() -> dict[str, Any]:  # type: ignore
        """Creates new custom fields for a Track.

        Returns:
            Dict of the field names to their default values or ``None`` for no default.

        Example:
            Inside your hook implementation::

                return {"my_new_field": "default value", "other_field": None}

            You can then access your new field as if it were a normal field::

                track.my_new_field = "awesome new value"

        Important:
            Your custom field should follow the same naming rules as any other python
            variable i.e. no spaces, starts with a letter, and consists solely of
            alpha-numeric and underscore characters.
        """  # noqa: DAR202

    @staticmethod
    @moe.hookspec
    def is_unique_track(track: "Track", other: "Track") -> bool:  # type: ignore
        """Add new conditions to determine whether two tracks are unique.

        "Uniqueness" is meant in terms of whether the two tracks should be considered
        duplicates in the library. These additional conditions will be applied inside a
        track's :meth:`is_unique` method.
        """

    @staticmethod
    @moe.hookspec
    def read_custom_tags(
        track_path: Path, album_fields: dict[str, Any], track_fields: dict[str, Any]
    ) -> None:
        """Read and set any fields from a track_path.

        How you read the file and assign tags is up to each individual plugin.
        Internally, Moe uses the `mediafile <https://github.com/beetbox/mediafile>`_
        library to read tags.

        Args:
            track_path: Path of the track file to read.
            album_fields: Dictionary of album fields to read from the given track file's
                tags. The dictionary may contain existing fields and values, and you
                can choose to either override the existing fields, or to provide new
                fields.
            track_fields: Dictionary of track fields to read from the given track file's
                tags. The dictionary may contain existing fields and values, and you
                can choose to either override the existing fields, or to provide new
                fields.

        Example:
            .. code:: python

                audio_file = mediafile.MediaFile(track_path)
                album_fields["title"] = audio_file.album
                track_fields["title"] = audio_file.title

        See Also:
            * :ref:`Album and track fields <fields:Fields>`
            * `Mediafile docs <https://mediafile.readthedocs.io/en/latest/>`_
            * The :meth:`~moe.plugins.write.Hooks.write_custom_tags` hook for writing
              tags.
        """


@moe.hookimpl
def add_hooks(pm: pluggy.manager.PluginManager):
    """Registers `track` hookspecs to Moe."""
    from moe.library.track import Hooks

    pm.add_hookspecs(Hooks)


@moe.hookimpl(tryfirst=True)
def read_custom_tags(
    track_path: Path, album_fields: dict[str, Any], track_fields: dict[str, Any]
) -> None:
    """Read and set internally tracked fields."""
    audio_file = mediafile.MediaFile(track_path)

    album_fields["artist"] = audio_file.albumartist or audio_file.artist
    album_fields["country"] = audio_file.country
    album_fields["date"] = audio_file.date
    album_fields["disc_total"] = audio_file.disctotal
    album_fields["label"] = audio_file.label
    album_fields["media"] = audio_file.media
    album_fields["original_date"] = audio_file.original_date
    album_fields["path"] = track_path.parent
    album_fields["title"] = audio_file.album

    track_fields["artist"] = audio_file.artist
    track_fields["artists"] = set(audio_file.artists)
    track_fields["disc"] = audio_file.disc
    track_fields["genres"] = set(audio_file.genres)
    track_fields["path"] = track_path
    track_fields["title"] = audio_file.title
    track_fields["track_num"] = audio_file.track


class TrackError(LibraryError):
    """Error performing some operation on a Track."""


class _Artist(SABase):
    """A track can have multiple artists."""

    __tablename__ = "artist"

    _id: int = cast(int, Column(Integer, primary_key=True))
    _track_id: int = cast(int, Column(Integer, ForeignKey("track._id")))
    name: str = cast(str, Column(String, nullable=False))

    def __init__(self, name: str):
        self.name = name


class _Genre(SABase):
    """A track can have multiple genres."""

    __tablename__ = "genre"

    _id: int = cast(int, Column(Integer, primary_key=True))
    _track_id: int = cast(int, Column(Integer, ForeignKey("track._id")))
    name: str = cast(str, Column(String, nullable=False))

    def __init__(self, name: str):
        self.name = name


# Track generic, used for typing classmethod
T = TypeVar("T", bound="Track")


class Track(LibItem, SABase):
    """A single track.

    Attributes:
        album (str)
        albumartist (str)
        album_obj (Album): Corresponding Album object.
        artist (str)
        artists (set[str]): Set of all artists.
        disc (int): Disc number the track is on.
        genre (str): String of all genres concatenated with ';'.
        genres (set[str]): Set of all genres.
        path (Path): Filesystem path of the track file.
        title (str)
        track_num (int)

    Note:
        Altering any album-related property attributes, will result in changing the
        album field and thus all other tracks in the album as well.
    """

    __tablename__ = "track"

    _id: int = cast(int, Column(Integer, primary_key=True))
    artist: str = cast(str, Column(String, nullable=False))
    disc: int = cast(int, Column(Integer, nullable=False, default=1))
    path: Path = cast(Path, Column(PathType, nullable=False, unique=True))
    title: str = cast(str, Column(String, nullable=False))
    track_num: int = cast(int, Column(Integer, nullable=False))
    _custom_fields: dict[str, Any] = cast(
        dict[str, Any],
        Column(
            MutableDict.as_mutable(JSON(none_as_null=True)),
            default="{}",
            nullable=False,
        ),
    )

    _album_id: int = cast(int, Column(Integer, ForeignKey("album._id")))
    album_obj: Album = relationship("Album", back_populates="tracks")
    album: str = association_proxy("album_obj", "title")
    albumartist: str = association_proxy("album_obj", "artist")

    _artists: set[_Artist] = relationship(
        "_Artist", collection_class=set, cascade="save-update, merge, expunge"
    )
    artists: set[str] = association_proxy("_artists", "name")
    _genres: set[_Genre] = relationship(
        "_Genre", collection_class=set, cascade="save-update, merge, expunge"
    )
    genres: set[str] = association_proxy("_genres", "name")

    __table_args__ = (UniqueConstraint("disc", "track_num", "_album_id"),)

    def __init__(
        self,
        album: Album,
        path: Path,
        title: str,
        track_num: int,
        **kwargs,
    ):
        """Creates a Track.

        Args:
            album: Album the track belongs to.
            path: Filesystem path of the track file.
            title: Title of the track.
            track_num: Track number.
            **kwargs: Any other fields to assign to the track.
        """
        self._custom_fields = self._get_default_custom_fields()
        self._custom_fields_set = set(self._custom_fields)

        album.tracks.append(self)
        self.path = path
        self.title = title
        self.track_num = track_num

        self.artist = self.albumartist  # default value

        for key, value in kwargs.items():
            setattr(self, key, value)

        if not self.disc:
            self.disc = self._guess_disc()

        log.debug(f"Track created. [track={self!r}]")

    def _guess_disc(self) -> int:
        """Attempts to guess the disc of a track based on it's path."""
        log.debug(f"Guessing track disc number. [track={self!r}]")

        if self.path.parent == self.album_obj.path:
            return 1

        # The track is in a subdirectory of the album - most likely disc directories.
        disc_dirs: list[Path] = []
        for path in self.album_obj.path.iterdir():
            if not path.is_dir():
                continue

            contains_tracks = False
            for album_track in self.album_obj.tracks:
                if album_track.path.is_relative_to(path):
                    contains_tracks = True

            if contains_tracks:
                disc_dirs.append(path)

        # Guess the disc by the order of the disc directory it appears in.
        for disc_num, disc_dir in enumerate(sorted(disc_dirs), start=1):
            if self.path.is_relative_to(disc_dir):
                return disc_num

        return 1

    @classmethod
    def from_file(cls: type[T], track_path: Path, album: Optional[Album] = None) -> T:
        """Alternate initializer that creates a Track from a track file.

        Will read any tags from the given path and save them to the Track.

        Args:
            track_path: Filesystem path of the track.
            album: Corresponding album for the track. If not given, the album will be
                created.

        Returns:
            Track instance.

        Raises:
            TrackError: Given ``path`` does not correspond to a track file.
        """
        log.debug(f"Creating track from path. [path={track_path}, {album=}]")

        try:
            mediafile.MediaFile(track_path)
        except mediafile.UnreadableFileError as err:
            raise TrackError(
                "Unable to create track; given path is not a track file. "
                f"[path={track_path}]"
            ) from err

        album_fields: dict[str, Any] = {}
        track_fields: dict[str, Any] = {}
        config.CONFIG.pm.hook.read_custom_tags(
            track_path=track_path, album_fields=album_fields, track_fields=track_fields
        )
        if not album:
            album = Album(
                path=album_fields.pop("path"),
                artist=album_fields.pop("artist"),
                title=album_fields.pop("title"),
                date=album_fields.pop("date"),
                disc_total=album_fields.pop("disc_total"),
                **album_fields,
            )

        return cls(
            album=album,
            path=track_fields.pop("path"),
            title=track_fields.pop("title"),
            track_num=track_fields.pop("track_num"),
            **track_fields,
        )

    @property
    def genre(self) -> str:
        """Returns a string of all genres concatenated with ';'."""
        return ";".join(self.genres)

    @genre.setter
    def genre(self, genre_str: str):
        """Sets a track's genre from a string.

        Args:
            genre_str: For more than one genre, they should be split with ';'.
        """
        self.genres = {genre.strip() for genre in genre_str.split(";")}

    @property
    def fields(self) -> set[str]:
        """Returns any editable, track-specific fields."""
        return {
            "album_obj",
            "artist",
            "artists",
            "disc",
            "genres",
            "path",
            "title",
            "track_num",
        }.union(set(self._custom_fields))

    def is_unique(self, other: "Track") -> bool:
        """Returns whether a track is unique in the library from ``other``."""
        if self.path == other.path:
            return False
        if (
            self.track_num == other.track_num
            and self.disc == other.disc
            and self.album_obj == other.album_obj
        ):
            return False

        custom_uniqueness = config.CONFIG.pm.hook.is_unique_track(
            track=self, other=other
        )
        if False in custom_uniqueness:
            return False

        return True

    def merge(self, other: "Track", overwrite: bool = False):
        """Merges another track into this one.

        Args:
            other: Other track to be merged with the current track.
            overwrite: Whether or not to overwrite self if a conflict exists.
        """
        log.debug(
            f"Merging tracks. [track_a={self!r}, track_b={other!r}, {overwrite=!r}]"
        )

        omit_fields = {"album_obj", "year"}
        for field in self.fields - omit_fields:
            other_value = getattr(other, field)
            self_value = getattr(self, field)
            if other_value and (overwrite or (not overwrite and not self_value)):
                setattr(self, field, other_value)

        log.debug(
            f"Tracks merged. [track_a={self!r}, track_b={other!r}, {overwrite=!r}]"
        )

    def __eq__(self, other) -> bool:
        """Compares Tracks by their fields."""
        if not isinstance(other, Track):
            return False

        for field in self.fields:
            if not hasattr(other, field) or (
                getattr(self, field) != getattr(other, field)
            ):
                return False

        return True

    def __lt__(self, other) -> bool:
        """Sort based on album, then disc, then track number."""
        if self.album_obj == other.album_obj:
            if self.disc == other.disc:
                return self.track_num < other.track_num

            return self.disc < other.disc

        return self.album_obj < other.album_obj

    def __repr__(self):
        """Represents a Track using track-specific and relevant album fields."""
        field_reprs = []
        omit_fields = {"album_obj"}
        for field in self.fields - omit_fields:
            if hasattr(self, field):
                field_reprs.append(f"{field}={getattr(self, field)!r}")
        repr_str = "Track(" + ", ".join(field_reprs) + f", album='{self.album_obj}'"

        custom_field_reprs = []
        for custom_field, value in self._custom_fields.items():
            custom_field_reprs.append(f"{custom_field}={value}")
        if custom_field_reprs:
            repr_str += ", custom_fields=[" + ", ".join(custom_field_reprs) + "]"

        repr_str += ")"
        return repr_str

    def __str__(self):
        """String representation of a track."""
        return f"{self.artist} - {self.title}"

    def _get_default_custom_fields(self) -> dict[str, Any]:
        """Returns the default custom track fields."""
        return {
            field: default_val
            for plugin_fields in config.CONFIG.pm.hook.create_custom_track_fields()
            for field, default_val in plugin_fields.items()
        }



"""Musicbrainz integration plugin.

See Also:
    * https://musicbrainz.org/doc/MusicBrainz_API/
    * https://python-musicbrainzngs.readthedocs.io/en/latest/api/
"""

import pluggy

import moe
from moe import config

from . import mb_cli, mb_core
from .mb_core import *

__all__ = []
__all__.extend(mb_core.__all__)


@moe.hookimpl
def plugin_registration():
    """Only register the cli sub-plugin if the cli is enabled."""
    config.CONFIG.pm.register(mb_core, "musicbrainz_core")
    if config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.register(mb_cli, "musicbrainz_cli")



"""Core api for removing items from the library."""

import logging

import sqlalchemy
import sqlalchemy.exc
from sqlalchemy.orm.session import Session

from moe.config import MoeSession
from moe.library import Extra, LibItem, Track

__all__ = ["remove_item"]

log = logging.getLogger("moe.remove")


def remove_item(item: LibItem):
    """Removes an item from the library."""
    log.debug(f"Removing item from the library. [{item=!r}]")
    session = MoeSession()

    insp = sqlalchemy.inspect(item)
    if insp.persistent:
        session.delete(item)
    elif insp.pending:
        session.expunge(item)
        if isinstance(item, (Track, Extra)):
            item.album_obj = None  # type: ignore

    try:
        session.flush()
    except sqlalchemy.exc.InvalidRequestError:
        # session is currently flushing - delete in separate session to ensure it
        # occurs before any inserts or updates in the original session
        if insp.persistent:
            session.expunge(item)
            new_session = Session(session.connection())
            new_session.delete(item)
            new_session.flush()

    log.info(f"Removed item from the library. [{item=!r}]")



"""Adds a duplicate resolution prompt to the CLI."""

import logging
from typing import Optional, Union, cast

from rich.columns import Columns
from rich.console import Group, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

import moe
import moe.cli
from moe.cli import console
from moe.library import Album, Extra, LibItem
from moe.plugins.remove import remove_item
from moe.util.cli import PromptChoice, choice_prompt

log = logging.getLogger("moe.cli.dup")


__all__: list[str] = []


@moe.hookimpl(trylast=True)
def resolve_dup_items(item_a: LibItem, item_b: LibItem):
    """Resolve any library duplicate conflicts using a user prompt."""
    console.print(_fmt_item_vs(item_a, item_b))

    # Each PromptChoice `func` should have the following signature:
    # func(item_a, item_b) # noqa: E800
    prompt_choices = [
        PromptChoice(title="Keep item A", shortcut_key="a", func=_keep_a),
        PromptChoice(title="Keep item B", shortcut_key="b", func=_keep_b),
        PromptChoice(
            title="Merge A -> B without overwriting conflicts",
            shortcut_key="m",
            func=_merge,
        ),
        PromptChoice(
            title="Merge A -> B, overwriting any conflicts",
            shortcut_key="o",
            func=_overwrite,
        ),
    ]
    prompt_choice = choice_prompt(
        prompt_choices,
        "Duplicate items found in the library, how would you like to resolve it?",
    )
    prompt_choice.func(item_a, item_b)


def _keep_a(item_a: LibItem, item_b: LibItem):
    """Keeps `item_a`, removing `item_b` from the library."""
    log.debug("Keeping item A.")

    remove_item(item_b)


def _keep_b(item_a: LibItem, item_b: LibItem):
    """Keeps `item_a`, removing `item_b` from the library."""
    log.debug("Keeping item B.")

    remove_item(item_a)


def _merge(item_a: LibItem, item_b: LibItem):
    """Merges `item_a` into `item_b` without overwriting any conflicts."""
    log.debug("Merging A -> B without overwriting any conflicts.")

    item_b.merge(item_a)
    remove_item(item_a)


def _overwrite(item_a: LibItem, item_b: LibItem):
    """Merges `item_a` into `item_b`, overwriting any conflicts."""
    log.debug("Merging A -> B, overwriting B on conflict.")

    item_b.merge(item_a, overwrite=True)
    remove_item(item_b)


def _fmt_item_vs(item_a: LibItem, item_b: LibItem) -> Columns:
    """Formats the differences between two items for output."""
    panel_a_style = "spring_green4"
    panel_b_style = "dark_red"

    if isinstance(item_a, Album):
        panel_a_title = "Album A"
        panel_b_title = "Album B"
    elif isinstance(item_a, Extra):
        panel_a_title = "Extra A"
        panel_b_title = "Extra B"
    else:
        panel_a_title = "Track A"
        panel_b_title = "Track B"

    panel_a = Panel(
        _fmt_item_text(item_a, item_b), title=panel_a_title, border_style=panel_a_style
    )
    panel_b = Panel(
        _fmt_item_text(item_b, item_a), title=panel_b_title, border_style=panel_b_style
    )
    return Columns((panel_a, panel_b))


def _fmt_item_text(
    item: LibItem, other: LibItem, include_header: bool = True
) -> RenderableType:
    """Formats a library item highlighting any differences from `other`."""
    if isinstance(item, Album):
        header = Text(f"{item}", justify="center")
        omit_fields = {"extras", "tracks", "year"}
    elif isinstance(item, Extra):
        header = Text(f"{item.rel_path}\n{item.album_obj}", justify="center")
        omit_fields = {"album_obj", "rel_path", "path"}
    else:
        header = Text(f"{item.title}\n{item.album_obj}", justify="center")
        omit_fields = {
            "album",
            "albumartist",
            "album_obj",
            "date",
            "disc_total",
            "genre",
            "year",
        }

    item_diff = Table.grid("field value", padding=(0, 1))
    item_has_diff = False
    for field in sorted(item.fields - omit_fields):
        field_diff = _fmt_field_vs(item, other, field)
        if field_diff:
            item_has_diff = True
            item_diff.add_row(Text(field, style="italic grey69"), field_diff)

    if isinstance(item, Album) and isinstance(other, Album):
        item_diff = Group(item_diff, "", _fmt_album_lists(item, other))
        item_has_diff = True

    if include_header and item_has_diff:
        return Group(header, "", item_diff)
    elif include_header and not item_has_diff:
        return header
    return item_diff


def _fmt_album_lists(album: Album, other: Album) -> Union[Group, Table]:
    """Formats the track and extra lists for an album highlighting diffs from other."""
    tracklist = Table.grid("track", padding=(0, 1))
    for track in sorted(album.tracks):
        tracklist.add_row(f"{track.disc}.{track.track_num} - {track.title}")

        other_track = other.get_track(track.track_num, track.disc)
        if other_track:
            track_diff = cast(
                Table, _fmt_item_text(track, other_track, include_header=False)
            )
            if len(track_diff.rows):
                tracklist.add_row(Padding(track_diff, (0, 4)))

    extralist = Table.grid("extra", padding=(0, 1))
    for extra in sorted(album.extras):
        extralist.add_row(f"{extra.rel_path}")

        other_extra = other.get_extra(extra.rel_path)
        if other_extra:
            extra_diff = cast(
                Table, _fmt_item_text(extra, other_extra, include_header=False)
            )
            if len(extra_diff.rows):
                extralist.add_row(Padding(extra_diff, (0, 4)))

    if album.extras:
        return Group(tracklist, "", extralist)
    return tracklist


def _fmt_field_vs(item: LibItem, other: LibItem, field: str) -> Optional[Text]:
    """Highlights field differences between two items.

    Args:
        item: Item to base comparisons off of.
        other: Other item to compare against.
        field: Field to compare.

    Returns:
        A rich 'Text' based on the following cases:

        1. `item.field` DNE:
           return `None`
        2. `item.field` == `other.field`:
           return `None`
        3. `item.field` != `other.field`:
           `item.field` will be returned in yellow
        4. `item.field` exists and `other.field` DNE:
           `item.field` will be returned in green
    """
    item_field = getattr(item, field, None)
    other_field = getattr(other, field, None)

    if not item_field or (item_field == other_field):
        return None
    if item_field != other_field:
        return Text(str(item_field), style="yellow")
    if item_field and not other_field:
        return Text(str(item_field), style="green")

    return None



"""This is a Python script that is run whenever the alembic migration tool is invoked.

At the very least, it contains instructions to configure and generate a SQLAlchemy
engine, procure a connection from that engine along with a transaction, and then
invoke the migration engine, using the connection as a source of database connectivity.

The env.py script is part of the generated environment so that the way migrations run
is entirely customizable. The exact specifics of how to connect are here, as well as
the specifics of how the migration environment are invoked. The script can be modified
so that multiple engines can be operated upon, custom arguments can be passed into the
migration environment, application-specific libraries and models can be loaded in
and made available.

Alembic includes a set of initialization templates which feature different varieties
of env.py for different use cases.
"""

from logging.config import fileConfig

from alembic import context
from moe.config import Config
from moe.library.lib_item import SABase

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# don't configure logger if Moe is running
if config.attributes.get("configure_logger", True):
    fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = SABase.metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a connection
    with the context.
    """
    connectable = config.attributes.get("connection", None)
    if not connectable:
        # only create Engine if we don't have a Connection from the outside
        moe_config = Config(init_db=False)
        moe_config._init_db(create_tables=False)
        connectable = moe_config.engine

    # When connectable is already a Connection object, calling
    # connect() gives us a *branched connection*.
    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata, render_as_batch=True
        )

        with context.begin_transaction():
            context.run_migrations()


run_migrations_online()



"""Alters the location of items in your library."""

import pluggy

import moe
from moe import config

from . import move_cli, move_core
from .move_core import *

__all__ = []
__all__.extend(move_core.__all__)


@moe.hookimpl
def plugin_registration():
    """Only register the cli sub-plugin if the cli is enabled."""
    config.CONFIG.pm.register(move_core, "move_core")
    if config.CONFIG.pm.has_plugin("cli"):
        config.CONFIG.pm.register(move_cli, "move_cli")



"""Core musicbrainz api.

The ``musicbrainz`` core plugin provides the following functionality:

* Imports metadata from musicbrainz when adding a track or album to the library.
* Optionally updates a musicbrainz collection when items are added or removed from the
  library.

Note:
    This plugin is enabled by default.

See Also:
    * https://musicbrainz.org/doc/MusicBrainz_API/
    * https://python-musicbrainzngs.readthedocs.io/en/latest/api/
"""

import datetime
import logging
from pathlib import Path
from typing import Any, Callable, Optional

import dynaconf
import mediafile
import musicbrainzngs
import pkg_resources

import moe
from moe import config
from moe.library import Album, LibItem, Track
from moe.plugins.moe_import.import_core import CandidateAlbum
from moe.util.core import match

__all__ = [
    "MBAuthError",
    "add_releases_to_collection",
    "get_album_by_id",
    "get_candidate_by_id",
    "get_track_by_id",
    "rm_releases_from_collection",
    "set_collection",
]

log = logging.getLogger("moe.mb")

MAX_SEARCH_LIMIT = 100
"""Max number of entries that can be returned from a search to musicbrainz. This limit
is enforced by their api."""


class MBAuthError(Exception):
    """Musicbrainz user authentication error."""


musicbrainzngs.set_useragent(
    "moe",
    pkg_resources.get_distribution("moe").version,
    contact="https://mrmoe.readthedocs.io/en/latest/index.html",
)

# information to include in the release api query
RELEASE_INCLUDES = [
    "aliases",
    "annotation",
    "artists",
    "artist-credits",
    "artist-rels",
    "labels",
    "media",
    "recordings",
    "recording-level-rels",
    "release-groups",
    "work-level-rels",
    "work-rels",
]


@moe.hookimpl
def add_config_validator(settings: dynaconf.base.LazySettings):
    """Validates musicbrainz plugin configuration settings."""
    login_required = False

    settings.validators.register(
        dynaconf.Validator(
            "musicbrainz.collection.auto_add",
            "musicbrainz.collection.auto_remove",
            default=False,
        )
    )

    if settings.get("musicbrainz.collection.auto_add", False) or settings.get(
        "musicbrainz.collection.auto_remove", False
    ):
        login_required = True
        settings.validators.register(
            dynaconf.Validator("musicbrainz.collection.collection_id", must_exist=True)
        )

    if login_required:
        settings.validators.register(
            dynaconf.Validator(
                "musicbrainz.username", "musicbrainz.password", must_exist=True
            )
        )


@moe.hookimpl
def create_custom_album_fields() -> dict[str, Any]:  # type: ignore
    """Adds relevant musicbrainz fields to an album."""
    return {"mb_album_id": None}


@moe.hookimpl
def create_custom_track_fields() -> dict[str, Any]:  # type: ignore
    """Adds relevant musicbrainz fields to a track."""
    return {"mb_track_id": None}


@moe.hookimpl
def get_candidates(album: Album) -> list[CandidateAlbum]:
    """Applies musicbrainz metadata changes to a given album.

    Args:
        album: Original album used to search musicbrainz for a matching album.

    Returns:
        A new album containing all the corrected metadata from musicbrainz. Note, this
        album is not complete, as it will not contain any references to the filesystem.
    """
    log.debug(f"Getting candidate albums from musicbrainz. [{album=!r}")

    search_criteria: dict = {}
    search_criteria["artist"] = album.artist
    search_criteria["release"] = album.title
    search_criteria["date"] = album.date.isoformat()

    releases = musicbrainzngs.search_releases(limit=5, **search_criteria)

    candidates = []
    for release in releases["release-list"]:
        candidates.append(get_candidate_by_id(album, release["id"]))

    if not candidates:
        log.warning("No candidate albums found.")
    else:
        log.info(f"Found candidate albums. [{candidates=!r}]")

    return candidates


@moe.hookimpl
def process_removed_items(items: list[LibItem]):
    """Removes a release from a collection when removed from the library."""
    if not config.CONFIG.settings.musicbrainz.collection.auto_remove:
        return

    mb_ids = []
    for item in items:
        if isinstance(item, Album) and item.mb_album_id:
            mb_ids.append(item.mb_album_id)

    if mb_ids:
        try:
            rm_releases_from_collection(set(mb_ids))
        except MBAuthError as err:
            log.error(err)


@moe.hookimpl
def process_new_items(items: list[LibItem]):
    """Updates a user collection in musicbrainz with new releases."""
    if not config.CONFIG.settings.musicbrainz.collection.auto_add:
        return

    releases = set()
    for item in items:
        if isinstance(item, Album) and item.mb_album_id:
            releases.add(item.mb_album_id)

    if releases:
        try:
            add_releases_to_collection(releases)
        except MBAuthError as err:
            log.error(err)


@moe.hookimpl
def read_custom_tags(
    track_path: Path, album_fields: dict[str, Any], track_fields: dict[str, Any]
) -> None:
    """Read and set musicbrainz release IDs from a track file."""
    audio_file = mediafile.MediaFile(track_path)

    album_fields["mb_album_id"] = audio_file.mb_albumid
    track_fields["mb_track_id"] = audio_file.mb_releasetrackid


@moe.hookimpl
def sync_metadata(item: LibItem):
    """Sync musibrainz metadata for associated items."""
    if isinstance(item, Album) and hasattr(item, "mb_album_id"):
        item.merge(get_album_by_id(item.mb_album_id), overwrite=True)
    elif isinstance(item, Track) and hasattr(item, "mb_track_id"):
        item.merge(
            get_track_by_id(item.mb_track_id, item.album_obj.mb_album_id),
            overwrite=True,
        )


@moe.hookimpl
def write_custom_tags(track: Track):
    """Write musicbrainz ID fields as tags."""
    audio_file = mediafile.MediaFile(track.path)

    audio_file.mb_albumid = track.album_obj.mb_album_id
    audio_file.mb_releasetrackid = track.mb_track_id

    audio_file.save()


def add_releases_to_collection(
    releases: set[str], collection: Optional[str] = None
) -> None:
    """Adds releases to a musicbrainz collection.

    Args:
        releases: Musicbrainz release IDs to add to the collection.
        collection: Musicbrainz collection ID to add the releases to.
            If not given, defaults to the ``musicbrainz.collection.collection_id``
            config option.

    Raises:
        MBAuthError: Invalid musicbrainz user credentials in the configuration.
    """
    collection = (
        collection or config.CONFIG.settings.musicbrainz.collection.collection_id
    )

    log.debug(
        f"Adding releases to musicbrainz collection. [{releases=!r}, {collection=!r}]"
    )

    _mb_auth_call(
        musicbrainzngs.add_releases_to_collection,
        collection=collection,
        releases=releases,
    )

    log.info(
        f"Added releases to musicbrainz collection. [{releases=!r}, {collection=!r}]"
    )


def rm_releases_from_collection(
    releases: set[str], collection: Optional[str] = None
) -> None:
    """Removes releases from a musicbrainz collection.

    Args:
        releases: Musicbrainz release IDs to remove from the collection.
        collection: Musicbrainz collection ID to remove the releases from.
            If not given, defaults to the ``musicbrainz.collection.collection_id``
            config option.

    Raises:
        MBAuthError: Invalid musicbrainz user credentials in the configuration.
    """
    collection = (
        collection or config.CONFIG.settings.musicbrainz.collection.collection_id
    )

    log.debug(
        "Removing releases from musicbrainz collection. "
        f"[{releases=!r}, {collection=!r}]"
    )

    _mb_auth_call(
        musicbrainzngs.remove_releases_from_collection,
        collection=collection,
        releases=releases,
    )

    log.info(
        "Removed releases from musicbrainz collection. "
        f"[{releases=!r}, {collection=!r}]"
    )


def _mb_auth_call(api_func: Callable, **kwargs) -> Any:
    """Call a musicbrainz API function that requires user authentication.

    Args:
        api_func: Musicbrainz API function to call.
        **kwargs: Keyword arguments to pass to the API function call.

    Returns:
        The return value of the API function called.

    Raises:
        MBAuthError: Invalid user credentials in the configuration.
    """
    musicbrainzngs.auth(
        config.CONFIG.settings.musicbrainz.username,
        config.CONFIG.settings.musicbrainz.password,
    )

    try:
        return api_func(**kwargs)
    except musicbrainzngs.AuthenticationError as err:
        raise MBAuthError("User authentication with musicbrainz failed.") from err


def set_collection(releases: set[str], collection: Optional[str] = None) -> None:
    """Sets a musicbrainz collection with the given releases.

    The releases in the collection will be set to ``releases``, adding any releases not
    present in the collection, as well as removing any extraneous releases.

    Args:
        releases: Musicbrainz releases to set the collection to.
        collection: Musicbrainz collection ID for the collection to set.
            If not given, defaults to the ``musicbrainz.collection.collection_id``
            config option.

    Raises:
        MBAuthError: Invalid user credentials in the configuration.
    """
    collection = (
        collection or config.CONFIG.settings.musicbrainz.collection.collection_id
    )

    log.debug("Setting musicbrainz collection. " f"[{releases=!r}, {collection=!r}]")

    current_releases = []
    num_searches = 0
    limit = MAX_SEARCH_LIMIT
    while len(current_releases) == limit * num_searches:  # hitting the search limit
        result = _mb_auth_call(
            musicbrainzngs.get_releases_in_collection,
            collection=collection,
            limit=limit,
            offset=limit * num_searches,
        )
        for release in result["collection"]["release-list"]:
            current_releases.append(release["id"])

        num_searches += 1
    current_releases = set(current_releases)

    stale_releases = current_releases.difference(releases)
    rm_releases_from_collection(stale_releases, collection)

    new_releases = releases.difference(current_releases)
    add_releases_to_collection(new_releases, collection)


def get_album_by_id(release_id: str) -> Album:
    """Returns an album from musicbrainz with the given release ID."""
    log.debug(f"Fetching release from musicbrainz. [release={release_id!r}]")

    release = musicbrainzngs.get_release_by_id(release_id, includes=RELEASE_INCLUDES)

    log.info(f"Fetched release from musicbrainz. [release={release_id!r}]")

    return _create_album(release["release"])


def get_candidate_by_id(album: Album, release_id: str) -> CandidateAlbum:
    """Returns a candidate for ``album`` from the given ``release_id``."""
    log.debug(f"Fetching release from musicbrainz. [release={release_id!r}]")
    release = musicbrainzngs.get_release_by_id(release_id, includes=RELEASE_INCLUDES)[
        "release"
    ]
    log.info(f"Fetched release from musicbrainz. [release={release_id!r}]")

    candidate_album = _create_album(release)

    sub_header_info = [str(_parse_date(release["date"]).year)]
    if disambiguation := release.get("disambiguation"):
        sub_header_info.append(disambiguation)

    return CandidateAlbum(
        album=candidate_album,
        match_value=match.get_match_value(album, candidate_album),
        source_str=f"Musicbrainz: {release_id}",
        sub_header_info=sub_header_info,
    )


def _create_album(release: dict) -> Album:
    """Creates an album from a given musicbrainz release."""
    log.debug(f"Creating album from musicbrainz release. [release={release['id']!r}]")

    if release["label-info-list"]:
        label = release["label-info-list"][0]["label"]["name"]
    else:
        label = None

    album = Album(
        artist=_flatten_artist_credit(release["artist-credit"]),
        country=release.get("country"),
        date=_parse_date(release["date"]),
        disc_total=int(release["medium-count"]),
        label=label,
        mb_album_id=release["id"],
        media=release["medium-list"][0]["format"],
        original_date=_parse_date(release["release-group"]["first-release-date"]),
        title=release["title"],
        path=None,  # type: ignore # this will get set in `add_prompt`
    )
    for medium in release["medium-list"]:
        for track in medium["track-list"]:
            Track(
                album=album,
                track_num=int(track["position"]),
                path=None,  # type: ignore # this will get set in `add_prompt`
                artist=_flatten_artist_credit(track["recording"]["artist-credit"]),
                disc=int(medium["position"]),
                mb_track_id=track["id"],
                title=track["recording"]["title"],
            )

    log.debug(f"Created album from musicbrainz release. [{album=!r}]")
    return album


def _flatten_artist_credit(artist_credit: list[dict]) -> str:
    """Given a musicbrainz formatted artist-credit, return the full artist name."""
    full_artist = ""
    for artist in artist_credit:
        if isinstance(artist, str):
            full_artist += artist
        else:
            full_artist += artist["artist"]["name"]

    return full_artist


def _parse_date(date: str) -> datetime.date:
    """Parses a date from a musicbrainz release."""
    date_parts = date.split("-")

    year = int(date_parts[0])
    try:
        month = int(date_parts[1])
    except IndexError:
        month = 1
    try:
        day = int(date_parts[2])
    except IndexError:
        day = 1

    return datetime.date(year, month, day)


def get_track_by_id(track_id: str, album_id: str) -> Track:
    """Gets a musicbrainz track from a given track and release id.

    Args:
        track_id: Musicbrainz track ID to match.
        album_id: Release album ID the track belongs to.

    Returns:
        A track containing all metadata associated with the given IDs.

    Raises:
        ValueError: Couldn't find track based on given ``track_id`` and ``album_id``.
    """
    log.debug(f"Fetching track from musicbrainz. [{track_id=!r}, {album_id=!r}]")

    album = get_album_by_id(album_id)
    for track in album.tracks:
        if track.mb_track_id == track_id:
            log.info(f"Fetched track from musicbrainz. [{track=!r}]")
            return track

    raise ValueError(
        f"Given track or album id could not be found. [{track_id=!r}, {album_id=!r}]"
    )



"""Adds the ``edit`` command to the cli."""

import argparse
import logging

import moe
import moe.cli
from moe.plugins import edit
from moe.util.cli import cli_query, query_parser

log = logging.getLogger("moe.cli.edit")


@moe.hookimpl
def add_command(cmd_parsers: argparse._SubParsersAction):
    """Adds the ``edit`` command to Moe's CLI."""
    epilog_help = """
    The FIELD=VALUE argument sets a track's field, an album's field if an album query is
    specified with `-a`, or an extra's field if an extra query is specified with `-e`.
    If the specified field supports multiple values, then you can separate those values
    with a semicolon. For example, `genre=hip hop;pop`.

    For more information on editing, see the documentation
    https://mrmoe.readthedocs.io/en/latest/plugins/edit.html
    """

    edit_parser = cmd_parsers.add_parser(
        "edit",
        description="Edits music in the library.",
        help="edit music in the library",
        parents=[query_parser],
        epilog=epilog_help,
    )
    edit_parser.add_argument(
        "fv_terms", metavar="FIELD=VALUE", nargs="+", help="set FIELD to VALUE"
    )
    edit_parser.set_defaults(func=_parse_args)


def _parse_args(args: argparse.Namespace):  # noqa: C901
    """Parses the given commandline arguments.

    Args:
        args: Commandline arguments to parse.

    Raises:
        SystemExit: Invalid query, no items found to edit, or invalid field or
            field_value term format.
    """
    items = cli_query(args.query, args.query_type)

    error_count = 0
    for term in args.fv_terms:
        try:
            field, value = term.split("=")
        except ValueError:
            log.error(f"Invalid FIELD=VALUE format. [{term=!r}]")
            error_count += 1
            continue

        for item in items:
            try:
                edit.edit_item(item, field, value)
            except edit.EditError as err:
                log.error(err)
                error_count += 1

    if error_count:
        raise SystemExit(1)



"""initial generation.

Revision ID: 6d4e785df5cb
Revises: 
Create Date: 2022-09-27 19:19:39.281403

"""
import sqlalchemy as sa
from sqlalchemy.ext.mutable import MutableDict

import moe
from alembic import op

# revision identifiers, used by Alembic.
revision = "6d4e785df5cb"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "album",
        sa.Column("_id", sa.Integer(), nullable=False),
        sa.Column("artist", sa.String(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("disc_total", sa.Integer(), nullable=False, default=1),
        sa.Column("path", moe.library.lib_item.PathType(), nullable=False),
        sa.Column("title", sa.String(), nullable=False),
        sa.Column(
            "_custom_fields",
            MutableDict.as_mutable(sa.JSON(none_as_null=True)),
            nullable=False,
            default={},
        ),
        sa.PrimaryKeyConstraint("_id"),
        sa.UniqueConstraint("path"),
    )
    op.create_table(
        "extra",
        sa.Column("_id", sa.Integer(), nullable=False),
        sa.Column("path", moe.library.lib_item.PathType(), nullable=False),
        sa.Column(
            "_custom_fields",
            MutableDict.as_mutable(sa.JSON(none_as_null=True)),
            nullable=False,
            default={},
        ),
        sa.Column("_album_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["_album_id"],
            ["album._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
        sa.UniqueConstraint("path"),
    )
    op.create_table(
        "track",
        sa.Column("_id", sa.Integer(), nullable=False),
        sa.Column("artist", sa.String(), nullable=False),
        sa.Column("disc", sa.Integer(), nullable=False),
        sa.Column("path", moe.library.lib_item.PathType(), nullable=False),
        sa.Column("title", sa.String(), nullable=False),
        sa.Column("track_num", sa.Integer(), nullable=False),
        sa.Column(
            "_custom_fields",
            MutableDict.as_mutable(sa.JSON(none_as_null=True)),
            nullable=False,
            default={},
        ),
        sa.Column("_album_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["_album_id"],
            ["album._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
        sa.UniqueConstraint("disc", "track_num", "_album_id"),
        sa.UniqueConstraint("path"),
    )
    op.create_table(
        "genre",
        sa.Column("_id", sa.Integer(), nullable=False),
        sa.Column("_track_id", sa.Integer(), nullable=True),
        sa.Column("name", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["_track_id"],
            ["track._id"],
        ),
        sa.PrimaryKeyConstraint("_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("genre")
    op.drop_table("track")
    op.drop_table("extra")
    op.drop_table("album")
    # ### end Alembic commands ###
